//  RW Dolldrop
// ==============
// External File Transfers Library
// -Part of the RW Foundations suite of libraries

// Filename: ./include/dollhouse/net/dolldrop.hpp
// -Declarations of the baseline RW Dolldrop sub-library

#pragma once

///////
// Dolldrop/Dollnet is meant for network-enabled RW projects only
#if __has_include(<curl_easy.h>)

#include "incls_all.hpp"

#include "curl_easy.h"

#include "dollnet_util.hpp"

//////////////////
//  'Net Work :
//

// DESIGN:
// -As a simple, basic control mechanism for starters we'll use some
// protected/private classes and a few functions from here within Dolldrop.
// Later, we can add more access restrictions and some further data validations;
// improving things bit-by-bit as we go along, using just a few of the basic &
// straightforward language features readily available to us from the standard.

// DESIGN:
// - add work-stealing queues of the Dolldrops/Mailboxes ?
// - use S.Parent's libdispatch-based async
// - this is actually a general need, so add it to the RW Foundations instead

// DESIGN:
// void push_drop() {}
// void yank_drop() {}

// DESIGN: TODO:
// return do_atomiconn(
//   lockout();
//     connect();
//       do_req(uri);
//     disconnect();
//   unlockout();
// );

// DESIGN: during this prototyping phase, code-envy is acceptable. however, for
// the production code abstract common behaviours into modular functions.
// -DRY principle

namespace rw::dollhouse::net {

using curl::curl_easy;

//------------------------------------------------------------------------------
/**
 */
class Dolldrop_b : public RW_slave {
 public:
  Dolldrop_b(Hip const& tag, std::string const& handle);  // tag-parm'd ctor

 protected:
  /** Perform a 'checked-pull' of a dolldrop, saving the transfer out to the
   * specified filepath @a droppath. It reuses a @c curl connection @a conn for
   * this site.
   *
   * Attempts an external network connection with, and download transfer from,
   * @a uri. If successful, transfer's body stream for @a uri is saved out to
   * disk file (to location specified by @a droppath).
   *
   * @param[in] tag Master's HIP
   * @param[in/out] conn curl-reuse connection
   * @param[in] uri the URI to attempt a download connection with
   * @param[in] droppath the disk file location to save transfer body to
   * @return the handle to the filesystem disk file
   * @note: this function marshals both @a pull_hdr_g() & @a pull_body_g(). the
   * header is examined against the local data before the body pull occurs (if
   * needed).
   */
  auto chkd_pull_g(Hip const&         tag,
                   curl_easy&         conn,
                   std::string const& uri,
                   fs::path const&    droppath) -> std::fstream;

  /** Attempts an external network connection with, and download transfer from,
   * @a uri. If successful, transfer's header stream is returned.
   *
   * @param[in] tag Master's HIP
   * @param[in/out] conn curl-reuse connection
   * @param[in] uri the URI to attempt a download connection with
   * @return the header stream
   * @note: reuses a @c curl connection @a conn determined for this site.
   * @note: effectively a wrapper for the @c Dollnet member func @a dl_hdr()
   */
  auto pull_hdr_g(Hip const&         tag,
                  curl_easy&         conn,
                  std::string const& uri,
                  bool const         show_prog = true) -> std::stringstream;

  /** Use @c curl-multi to pull requested headers in parallel
   * @param[in] tag Master's HIP
   * @param[in] pull_list the source pull list to check
   * @return the now-populated string container of pulled headers
   */
  auto pull_hdrs_multi_g(Hip const& tag, Pull_list const& pull_list)
      -> std::vector<std::string>;

  /** Check and pull all headers that need updating, returning the updates list
   * @param[in] tag Master's HIP
   * @param[in] pull_list the source pull list to check
   * @return the list needing updates
   */
  auto chk_hdrs_g(Hip const&                      tag,
                  Pull_list const&                pull_list,
                  std::vector<std::string> const& hdrs_dat) -> Pull_list;

  /** Attempts an external network connection with, and download transfer from,
   * @a uri. If successful, transfer's body stream for @a uri is saved out to
   * disk file (to location specified by @a droppath).
   *
   * @param[in] tag Master's HIP
   * @param[in/out] conn curl-reuse connection
   * @param[in] uri the URI to attempt a download connection with
   * @param[in] droppath the disk file location to save transfer body to
   * @return the handle to the filesystem disk file
   * @note: reuses a @c curl connection @a conn determined for this site.
   * @note: effectively a wrapper for the @c Dollnet member func @a dl_body()
   */
  auto pull_body_g(Hip const&         tag,
                   curl_easy&         conn,
                   std::string const& uri,
                   fs::path const&    droppath) -> std::fstream;

  /** Attempts an external network connection with, and download transfer from,
   * @a uri. If successful, transfer's body stream for @a uri is saved out to
   * disk file (as a generic 'anons_file' form), and transfer's header stream is
   * returned.
   *
   * @param[in] tag Master's HIP
   * @param[in/out] conn curl-reuse connection
   * @param[in] uri the URI to attempt a download connection with
   * @return the header stream
   * @note: reuses a @c curl connection @a conn determined for this site.
   */
  auto pull_hdr_body_g(Hip const& tag, curl_easy& conn, std::string const& uri)
      -> std::stringstream;

  // DESIGN: consolidate this up into pull_hdr_body_g() ?
  //
  /** Attempts an external network connection with, and download transfer from,
   * @a uri. If successful, transfer's body stream for @a uri is saved out to
   * disk file (as a generic 'anons_file' form), and transfer's header stream is
   * returned.
   *
   * @param[in] tag Master's HIP
   * @param[in/out] conn curl-reuse connection
   * @param[in] uri the URI to attempt a download connection with
   * @return the header stream
   * @note: this function auto-generates the 'anons_file' filename
   * @note: reuses a @c curl connection @a conn determined for this site.
   * @note: effectively a wrapper for the @c Dollnet member func @a
   * dl_hdr_body()
   */
  auto try_hdr_body(Hip const& tag, curl_easy& conn, std::string const& uri)
      -> std::stringstream;

  /** Create a new @c curl easy object
   * @return the new curl
   * @note: simply a wrapper for same function in @c Dollnet member
   */
  auto get_curl() -> curl_easy;

  /** Retrieves (or creates a new) @c curl object, particular to a URI @a host
   * @param[in] uri the URI to match with a curl obj
   * @return the matched curl obj
   */
  auto get_ib_curl(std::string const& uri) -> curl_easy;

  /** Ensure the dolldrops directory is in place to begin with
   * @return true on successful creation
   */
  auto mk_dropdir() -> bool;

  /** Write a general usage notes text file to dolldrops directory
   */
  void wrt_tailme();

  /** Stores a given site's associated curl connection object; for connection
   * re-use
   */
  map<string, curl_easy> site_conns_;

  /** Signals a catalog pull request
   */
  bool is_catalog_{false};

 private:
  Dolldrop_b() = delete;  // no default ctor

  /** Initialize ./dolldrops/ directory
   */
  void preflight();

  //----------------------------------------------------------------------------
  /**
   */
  class Dollnet : public RW_slave {
   public:
    Dollnet(Hip const& tag, std::string const& handle);  // tag-parm'd ctor

    /** Attempts an external network connection with, and download transfer
     * from, @a uri. If successful, transfer's header stream is returned.
     *
     * @param[in] tag Master's HIP
     * @param[in/out] conn curl-reuse connection
     * @param[in] uri the URI to attempt a download connection with
     * @return the header stream
     * @note: reuses a @c curl connection @a conn determined for this site.
     */
    auto dl_hdr(Hip const&         tag,
                curl_easy&         conn,
                std::string const& uri,
                bool const         show_prog = true) -> std::stringstream;

    /** Attempts an external network connection with, and download transfer
     * from, @a uri. If successful, transfer's body stream for @a uri is saved
     * out to disk file @a dropfile.
     *
     * @param[in] tag Master's HIP
     * @param[in/out] conn curl-reuse connection
     * @param[in] uri the URI to attempt a download connection with
     * @param[in] dropfile the disk file to save transfer body to
     * @note: reuses a @c curl connection @a conn determined for this site.
     */
    void dl_body(Hip const&         tag,
                 curl_easy&         conn,
                 std::string const& uri,
                 std::ofstream&     dropfile);

    /** Attempts an external network connection with, and download transfer
     * from, @a uri. If successful, transfer's body stream for @a uri is saved
     * out to disk file @a dropfile, and transfer's header stream is returned.
     *
     * @param[in] tag Master's HIP
     * @param[in/out] conn curl-reuse connection
     * @param[in] uri the URI to attempt a download connection with
     * @param[in] dropfile the disk file to save transfer body to
     * @return the header stream
     * @note: reuses a @c curl connection @a conn determined for this site.
     */
    auto dl_hdr_body(Hip const&         tag,
                     curl_easy&         conn,
                     std::string const& uri,
                     std::ofstream&     dropfile) -> std::stringstream;

    /** Create a new @c curl easy object
     * @return the new curl
     */
    auto get_curl() -> curl_easy;

   private:
    Dollnet() = delete;  // no default ctor

  } dollnet_;

};  // class Dolldrop_b

//------------------------------------------------------------------------------
/**
 */
class Dolldrop : public Dolldrop_b {
 public:
  // NOTE: no ctor; class interface only

  /** Attempts an external network connection with, and download transfer from,
   * @a uri. If successful, transfer's body stream for @a uri is saved out to
   * disk file (to location specified in @a droppath)
   *
   * @param[in] uri the URI to attempt a download connection with
   * @param[in] droppath the disk file location to save transfer body to
   * @return parsed JSON from @a droppath disk file
   * @note: public wrapper for Master's use only
   */
  auto pull_to_path(std::string const& uri, fs::path const& droppath) -> Json;

  /** Attempts an external network connection with, and download transfer from,
   * @a uri. If successful, transfer's body stream for @a uri is saved out to
   * disk file (as a generic 'anons_file' form)
   *
   * @param[in] uri the URI to attempt a download connection with
   * @return header stream data for @a uri, if present
   * @note: public wrapper for Master's use only
   */
  auto pull_hdr_body(std::string const& uri) -> std::stringstream;

  /**
   * @param pull_list
   * @return
   */
  auto chk_hdrs(Pull_list const& pull_list) -> Pull_list;

  /** Assign to the *catalog pull request* signal
   * @param[in] flag
   * @note: controls state of protected member field in @c Dolldrop_b
   */
  void do_catalog(bool const flag);

 private:
  Dolldrop() = delete;  // no default ctor

};  // class Dolldrop

}  // namespace rw::dollhouse::net

#endif  // __has_include(<curl_easy.h>)

//    --=^=--

// "______"

// rw_dollhouse_lib.v220521
// Copyright (2022)
// License MIT (Expat) https://opensource.org/licenses/MIT
