//  RW Dolldrop
// ==============
// External File Transfers Library
// -Part of the RW Foundations suite of libraries

// Filename: ./src/dollhouse/net/dolldrop.cpp
// -Definitions of the baseline RW Dolldrop sub-library

///////
// Dolldrop/Dollnet is meant for network-enabled RW projects only
#if __has_include(<curl_easy.h>)

#include "incls_all.hpp"

#include <curl/multi.h>

#include "dolldrop_util.cpp"

namespace rw::dollhouse::net {

//------------------------------------------------------------------------------
Dolldrop_b::Dolldrop_b(Hip const& tag, std::string const& handle)
    : RW_slave{{Type::Bumpmaster, Type::Curator}, handle, Type::Dolldrop},
      dollnet_{t(), "dolldrop_dollnet"} {
  stamp("Dolldrop_b()");

  bind_master(tag);

  preflight();
}

//------------------------------------------------------------------------------
auto Dolldrop_b::chkd_pull_g(Hip const&         tag,
                             curl_easy&         conn,
                             std::string const& uri,
                             fs::path const&    droppath) -> std::fstream {
  stamp(tag, "chkd_pull_g()");

  try {
    // pull the dolldrop's header
    auto hdr = pull_hdr_g(tag, conn, uri);

    // check if the remote source (claims to be) modified since last download
    // DESIGN: 'name' now unused, drop from called func ret
    //
    const auto& [have_new_drop, name] = chk_droptimes(uri, hdr);

    if (have_new_drop) {  // yes, remote 'last-modified' header tag differs

      // check if the remote source's and local dolldrop's 'length' tags match
      const auto size_match = chk_dropsizes(uri, hdr);

      // -if this is a catalog's pull check, just a newer remote file is good
      // enough (since it's size may stay identical, even w/ a legitimate
      // change)
      // -otherwise, also require that the size of the remote differs from local
      //
      if (is_catalog_ || (! size_match)) {  // yes, do request the dolldrop pull

        // adjust the display cursor back up one line -re: prelim header request
        cout << "\u008D\r" << flush;

        return pull_body_g(tag, conn, uri, droppath);
      }
    }

    // otherwise, just return the already-saved (and current) local dolldrop
    return fstream{droppath};

  } catch (runtime_error const& e) {
    cerr << "\n  runtime error caught in chkd_pull_g()\n\n";
    return fstream{};

  } catch (...) {
    cerr << "\n  unknown error caught in chkd_pull_g()\n\n";
    return fstream{};
  }
}

//------------------------------------------------------------------------------
auto Dolldrop_b::pull_hdr_g(Hip const&         tag,
                            curl_easy&         conn,
                            std::string const& uri,
                            bool const         show_prog) -> std::stringstream {
  stamp(tag, "pull_hdr_g()");

  if (tag == master()) {  // function is meant only for slave's Master to use

    try {
      return dollnet_.dl_hdr(t(), conn, uri, show_prog);

    } catch (runtime_error const& e) {
      cerr << "\n  runtime error caught in pull_hdr_g()\n\n";
      return stringstream{};

    } catch (...) {
      cerr << "\n  unknown error caught in pull_hdr_g()\n\n";
      return stringstream{};
    }

  } else
    only_master_can(master(), tag, t(), "Dolldrop_b::pull_hdr_g()");

  ///
  return stringstream{};  // shouldn't reach here
}

//------------------------------------------------------------------------------
auto Dolldrop_b::pull_body_g(Hip const&         tag,
                             curl_easy&         conn,
                             std::string const& uri,
                             fs::path const&    droppath) -> std::fstream {
  stamp(tag, "pull_body_g()");

  if (tag == master()) {  // function is meant only for slave's Master to use

    set_name(uri, droppath);

    ofstream anons_file{droppath};

    try {
      dollnet_.dl_body(t(), conn, uri, anons_file);

    } catch (runtime_error const& e) {
      cerr << "\n  runtime error caught in pull_body_g()\n\n";
      anons_file.close();
      return fstream{};

    } catch (...) {
      cerr << "\n  unknown error caught in pull_body_g()\n\n";
      anons_file.close();
      return fstream{};
    }

    anons_file.close();

    // TEMP: return dolldrop body directly for now
    return fstream{droppath};

  } else
    only_master_can(master(), tag, t(), "Dolldrop_b::pull_body_g()");

  ///
  return fstream{};  // shouldn't reach here
}

//------------------------------------------------------------------------------
auto Dolldrop_b::pull_hdr_body_g(Hip const&         tag,
                                 curl_easy&         conn,
                                 std::string const& uri) -> std::stringstream {
  stamp(tag, "pull_hdr_body_g()");

  // TEST:
  prt_tag(tag, "asked to pull a dolldrop w/ hdr");
  cout << "  -requested pull address :  " << uri << "\n\n";

  if (tag == master()) {  // function is meant only for slave's Master to use
    return try_hdr_body(t(), conn, uri);

  } else
    only_master_can(master(), tag, t(), "Dolldrop_b::pull_hdr_body_g()");

  ///
  return stringstream{};  // shouldn't reach here
}

//------------------------------------------------------------------------------
auto Dolldrop_b::try_hdr_body(Hip const&         tag,
                              curl_easy&         conn,
                              std::string const& uri) -> std::stringstream {  //

  auto const   dropname = bld_dropname();
  ofstream     anons_file{dropname};
  stringstream hdr;

  try {
    hdr = dollnet_.dl_hdr_body(tag, conn, uri, anons_file);

  } catch (runtime_error const& e) {
    anons_file.close();
    set_name(uri, dropname);
    return stringstream{};

  } catch (...) {
    cerr << "\n  unknown error caught in try_hdr_body()\n\n";
    anons_file.close();
    set_name(uri, dropname);
    return stringstream{};
  }

  anons_file.close();
  set_name(uri, dropname);

  // TEMP: return dolldrop header directly for now
  return hdr;
}

//------------------------------------------------------------------------------
/** Append a @c curl -specified memory block to a target string
 *
 * Writes an array of @a num_elems count of elements, each one @a elem_sz in
 * size, from the block of memory pointed to by @a addr, out to the current
 * position in the @a o_strm stream (the end of a @c std::string, in this case).
 *
 * @param[in] addr pointer to memory array of elements to be written out
 * @param[in] elem_sz size in bytes of each element
 * @param[in] num_elems number of elements
 * @param[out] o_strm pointer to output stream
 * @return number of bytes written
 * @note: this is intended as a callback function for a curl handle's use
 */
auto wrt_str_cb(Address const      addr,
                size_t const       elem_sz,
                size_t const       num_elems,
                std::string* const o_strm) -> size_t {  //

  auto const mem_blk_sz{elem_sz * num_elems};

  try {
    o_strm->append(static_cast<const char*>(addr), mem_blk_sz);

  } catch (exception const&) {
    // -"If an exception is thrown for any reason, this function has no effect
    //  (strong exception guarantee). (since C++11)"
    return 0;
  }

  return mem_blk_sz;

  // https://en.cppreference.com/w/cpp/string/basic_string/append
}

//------------------------------------------------------------------------------
/** Assign curl easy options for a new @c curl handle, adding it into a @c
 * curl-multi handle
 *
 * @param[out] multi_hdl the multi that receives the new curl easy hdl
 * @param[in] uri the associated uri for this easy hdl
 * @param[out] hdrs_dat the target string container to receive header data
 * @param[in] idx index to the specific string to receive the header data
 */
void cfg_hdr_hdl(CURLM* const              multi_hdl,
                 std::string const&        uri,
                 std::vector<std::string>& hdrs_dat,
                 size_t const              idx) {
  auto curl_hdl = curl_easy_init();

  curl_easy_setopt(curl_hdl, CURLOPT_WRITEFUNCTION, wrt_str_cb);

  curl_easy_setopt(curl_hdl, CURLOPT_NOBODY, 1L);
  curl_easy_setopt(curl_hdl, CURLOPT_HEADERDATA, &hdrs_dat[idx]);

  curl_easy_setopt(curl_hdl, CURLOPT_USERAGENT, "Bumpmaster_v220521");

  curl_easy_setopt(curl_hdl, CURLOPT_URL, uri.c_str());
  curl_easy_setopt(curl_hdl, CURLOPT_FOLLOWLOCATION, 1L);

  curl_multi_add_handle(multi_hdl, curl_hdl);
}

//------------------------------------------------------------------------------
auto Dolldrop_b::pull_hdrs_multi_g(Hip const& tag, Pull_list const& pull_list)
    -> std::vector<std::string> {
  stamp(tag, "pull_hdrs_multi_g()");

  std::vector<std::string> hdrs_dat(pull_list.size());  // init w/ empties

  if (tag == master()) {  // function is meant only for slave's Master to use

    auto const chks_sz = pull_list.size();
    cout << "\n checking (" << chks_sz << ") item's headers..." << flush;

    auto multi_hdl = curl_multi_init();

    // load up the curl-multi handle, w/ each individual curl handle per header
    for (auto const& [uri, fs_path, idx] : pull_list)
      cfg_hdr_hdl(multi_hdl, uri, hdrs_dat, idx);

    auto           hdls_running{0};
    constexpr auto timeout_ms{20 * 1000};  // 20s (increase if needed, Anon)

    // perform the curl-multi operation, waiting for stragglers to complete
    //
    curl_multi_perform(multi_hdl, &hdls_running);
    do {
      auto ret{0};
      auto res_code = curl_multi_wait(multi_hdl, nullptr, 0, timeout_ms, &ret);

      if (res_code != CURLM_OK) {
        /// TBD...
      }

      curl_multi_perform(multi_hdl, &hdls_running);

    } while (hdls_running);

    // cleanups-only (messages ignored for now; TBD...)
    // -DESIGN: abstract this out if no further use comes up here
    //
    auto     msgs_in_queue{0};
    CURLMsg* msg = nullptr;
    //
    while ((msg = curl_multi_info_read(multi_hdl, &msgs_in_queue))) {
      auto curl_hdl = msg->easy_handle;
      curl_multi_remove_handle(multi_hdl, curl_hdl);
      curl_easy_cleanup(curl_hdl);
    }
    //
    curl_multi_cleanup(multi_hdl);

  } else
    only_master_can(master(), tag, t(), "Dolldrop_b::pull_hdrs_multi_g()");

  return hdrs_dat;
}

//------------------------------------------------------------------------------
auto Dolldrop_b::chk_hdrs_g(Hip const&                      tag,
                            Pull_list const&                pull_list,
                            std::vector<std::string> const& hdrs_dat)
    -> Pull_list {
  stamp(tag, "chk_hdrs_g()");

  Pull_list res;

  if (tag == master()) {  // function is meant only for slave's Master to use

    // check header strings contained in hdrs_dat, against the data in droplog
    //
    for (auto const& [uri, fs_path, idx] : pull_list) {
      stringstream hdr{hdrs_dat[idx]};  // an already-retrieved hdr for this uri

      // If header's DL failed for some reason, just attempt the curl DL anyway
      if (is_empty(hdr)) {
        res.push_back(make_tuple(uri, fs_path, idx));
        continue;  // skip further proc'ng for this dolldrop's header
      }

      // check if the remote source (claims to be) modified since last download
      // -DESIGN: 'name' now unused, drop from called func ret
      //
      const auto& [have_new_drop, name] = chk_droptimes(uri, hdr);

      if (have_new_drop) {  // yes, remote 'last-modified' header tag differs

        // check if the remote source's and local dolldrop's 'length' tags match
        const auto size_match = chk_dropsizes(uri, hdr);

        // -if this is a catalog's pull check, just a newer remote file is good
        // enough (since it's size may stay identical, even w/ a legitimate
        // change)
        // -otherwise, also require that the size of the remote differs from
        // local
        //
        if (is_catalog_ || (! size_match))  // yes, add item to pull list
          res.push_back(make_tuple(uri, fs_path, idx));
      }
    }

  } else
    only_master_can(master(), tag, t(), "Dolldrop_b::chk_hdrs_g()");

  return res;
}

//------------------------------------------------------------------------------
auto Dolldrop_b::get_curl() -> curl_easy {
  return dollnet_.get_curl();
}

//------------------------------------------------------------------------------
auto Dolldrop_b::get_ib_curl(std::string const& uri) -> curl_easy {
  auto const& [site_nm, board_nm] = parse_ib_uri(uri);

  // obtain the previous curl_easy obj for this site if found, else create new
  curl_easy conn;
  try {
    conn = site_conns_.at(site_nm);

  } catch (out_of_range const& e) {
    conn                 = get_curl();
    site_conns_[site_nm] = conn;
  }

  return conn;
}

//------------------------------------------------------------------------------
auto Dolldrop_b::mk_dropdir() -> bool {
  fs::create_directory(dropdir);
  return exists_fs(dropdir);
}

//------------------------------------------------------------------------------
void Dolldrop_b::wrt_tailme() {  //

  // a raw string; the body of the text file
  auto const tailtext =
      R"(//  RW Dolldrop
// ==============
// External File Transfers Library
// -Part of the RW Foundations suite of libraries

Notes:

-You can use the tail command to watch DLs progressively fill your dolldrops
 E.G., (in a terminal from the project's base directory) :

    tail -f  ./dolldrops/anons_file_001

-You can even playback a stream-capp style dolldrop 'live' (while its capturing)
 E.G. :

    vlc  ./dolldrops/anons_file_001
       or
    mpv  ./dolldrops/anons_file_042.ogg

-Protip : If you want a file extension automatically appended to your dolldrops,
 put it on the second line of the file named :  ./dolldrops/tagfile
 E.G., for OGG dolldrops, a tagfile might be edited to begin with the contents :

    000
    .ogg

 which would (on repeated use of this demo) create a series of dolldrops named :

    ./dolldrops/anons_file_001.ogg
    ./dolldrops/anons_file_002.ogg
    ./dolldrops/anons_file_003.ogg

 &tc. (btw this can be changed anytime to get a different extension next round)

Cheers.

// rw_dollhouse_lib
// Copyright (C)
// License MIT (Expat) https://opensource.org/licenses/MIT
)";

  ofstream ofs{dropdir / "tail_me.txt"};
  ofs << tailtext;
  ofs.close();
}

//------------------------------------------------------------------------------
void Dolldrop_b::preflight() {  //

  if (! exists_fs(dropdir)) {  // check if ./dolldrops/ directory exists

    try {
      mk_dropdir();
      wrt_tailme();

    } catch (fs::filesystem_error const& e) {
      cout << "  fs::filesystem_error : '" << e.what()
           << "' caught in Dolldrop_b::preflight()\n\n";

    } catch (std::bad_alloc const& e) {
      cout << "  std::bad_alloc : '" << e.what()
           << "' caught in Dolldrop_b::preflight()\n\n";
    }
  }
}

//------------------------------------------------------------------------------
auto Dolldrop::pull_to_path(std::string const& uri, fs::path const& droppath)
    -> Json {
  auto conn = get_ib_curl(uri);

  chkd_pull_g(master(), conn, uri, droppath);

  return rd_json_file(droppath);
}

//------------------------------------------------------------------------------
auto Dolldrop::pull_hdr_body(std::string const& uri) -> std::stringstream {
  auto conn = get_ib_curl(uri);

  return pull_hdr_body_g(master(), conn, uri);
}

//------------------------------------------------------------------------------
auto Dolldrop::chk_hdrs(Pull_list const& pull_list) -> Pull_list {  //

  curl_global_init(CURL_GLOBAL_ALL);

  auto const hdrs_dat = pull_hdrs_multi_g(master(), pull_list);

  return chk_hdrs_g(master(), pull_list, hdrs_dat);
}

//------------------------------------------------------------------------------
void Dolldrop::do_catalog(bool const flag) {
  stamp("do_catalog()");

  is_catalog_ = flag;
}

}  // namespace rw::dollhouse::net

#endif  // __has_include(<curl_easy.h>)

//    --=^=--

// "______"

// rw_dolldrop_lib.v220521
// Copyright (2022)
// License MIT (Expat) https://opensource.org/licenses/MIT
