//   RW Dollnet
// ==============
// External, Inter-Networking Library
// -Part of the RW Foundations suite of libraries

// Filename: ./src/dollhouse/net/dollnet.cpp
// -Definitions of the baseline RW Dollnet sub-library

///////
// Dolldrop/Dollnet is meant for network-enabled RW projects only
#if __has_include(<curl_easy.h>)

#include "incls_all.hpp"

#include "curl_easy.h"

#include "dollnet_util.cpp"

// DESIGN: during this prototyping phase, code-envy is acceptable. however, for
// the production code abstract common behaviours into modular functions.
// -DRY principle

namespace rw::dollhouse::net {

using curl::curl_easy;
using curl::curl_easy_exception;
using curl::curl_ios;

//------------------------------------------------------------------------------
/** Print the @a curl error's message then rethrow it as a @c std::runtime_error
 * @param e
 * @param func_nm
 */
void rprt_throw(curl_easy_exception const& e, std::string const& func_nm) {
  auto const& errors{e.get_traceback()};
  cerr << "\n  err in " << func_nm << '\n' << errors[0].first << '\n';

  throw runtime_error{e};
}

//------------------------------------------------------------------------------
Dolldrop_b::Dollnet::Dollnet(Hip const& tag, std::string const& handle)
    : RW_slave{{Type::Dolldrop}, handle} {
  stamp("Dollnet()");

  bind_master(tag);
}

//------------------------------------------------------------------------------
auto Dolldrop_b::Dollnet::dl_hdr(Hip const&         tag,
                                 curl_easy&         conn,
                                 std::string const& uri,
                                 bool const show_prog) -> std::stringstream {
  stamp(tag, "dl_hdr()");

  if (tag == master()) {  // function is meant only for slave's Master to use

    stringstream           header_dat{};
    curl_ios<stringstream> header{header_dat};
    conn.add<CURLOPT_HEADERDATA>(header.get_stream());

    conn.add<CURLOPT_NOBODY>(1L);

    if (show_prog) {
      conn.add<CURLOPT_XFERINFOFUNCTION>(prt_time);  // time only (no size)
      conn.add<CURLOPT_NOPROGRESS>(0L);
    }

    conn.add<CURLOPT_USERAGENT>("Bumpmaster_v220521");

    conn.add<CURLOPT_URL>(uri.c_str());
    conn.add<CURLOPT_FOLLOWLOCATION>(1L);

    try {
      start = Call_clk::now();
      conn.perform();

      if (show_prog)
        puts("");

    } catch (curl_easy_exception const& e) {
      rprt_throw(e, "Dollnet::dl_hdr()");
    }

    header_dat << flush;

    return header_dat;

  } else
    only_master_can(master(), tag, t(), "Dollnet::dl_hdr()");

  ///
  return stringstream{};  // shouldn't reach here
}

//------------------------------------------------------------------------------
void Dolldrop_b::Dollnet::dl_body(Hip const&         tag,
                                  curl_easy&         conn,
                                  std::string const& uri,
                                  std::ofstream&     dropfile) {
  stamp(tag, "dl_body()");

  if (tag == master()) {  // function is meant only for slave's Master to use

    conn.reset();  // necessary for now - DESIGN: any way to limit reset?

    curl_ios<ostream> writer{dropfile};
    conn.add<CURLOPT_WRITEFUNCTION>(writer.get_function());
    conn.add<CURLOPT_WRITEDATA>(writer.get_stream());

    conn.add<CURLOPT_HEADER>(0L);

    conn.add<CURLOPT_XFERINFOFUNCTION>(prt_prog);
    conn.add<CURLOPT_NOPROGRESS>(0L);

    conn.add<CURLOPT_USERAGENT>("Bumpmaster_v220521");

    conn.add<CURLOPT_URL>(uri.c_str());
    conn.add<CURLOPT_FOLLOWLOCATION>(1L);

    try {
      start = Call_clk::now();
      conn.perform();
      puts("");

    } catch (curl_easy_exception const& e) {
      rprt_throw(e, "Dollnet::dl_body()");
    }

  } else
    only_master_can(master(), tag, t(), "Dollnet::dl_body()");
}

//------------------------------------------------------------------------------
auto Dolldrop_b::Dollnet::dl_hdr_body(Hip const&         tag,
                                      curl_easy&         conn,
                                      std::string const& uri,
                                      std::ofstream&     dropfile)
    -> std::stringstream {
  stamp(tag, "dl_hdr_body()");

  if (tag == master()) {  // function is meant only for slave's Master to use

    conn.reset();  // necessary for now - DESIGN: any way to limit reset?

    // the "writer" stream will put the DL body's contents into "dropfile"
    curl_ios<ostream> writer{dropfile};

    // we'll use curlcpp's default write function for "writer"
    conn.add<CURLOPT_WRITEFUNCTION>(writer.get_function());

    // specify which stream (ie, write data) to use for "writer" 's contents
    conn.add<CURLOPT_WRITEDATA>(writer.get_stream());

    // we'll return this header data stream at the end of this code block
    stringstream header_dat{};

    // the "header" stream will put the DL header's contents into "header_dat"
    curl_ios<stringstream> header{header_dat};

    // specify which stream (ie, header data) to use for "header" 's contents
    conn.add<CURLOPT_HEADERDATA>(header.get_stream());

    // we'll output our DL progress to the console
    conn.add<CURLOPT_XFERINFOFUNCTION>(prt_prog);

    // yes, do go ahead and track our DL progress
    conn.add<CURLOPT_NOPROGRESS>(0L);

    // r/n we're simply known as 'Bumpmaster' to the rest of the world
    conn.add<CURLOPT_USERAGENT>("Bumpmaster_v220521");

    // here's where we're going
    conn.add<CURLOPT_URL>(uri.c_str());

    // yes, follow locations
    conn.add<CURLOPT_FOLLOWLOCATION>(1L);

    try {  // can catch exceptions for this curlcpp Easy object (which is great)

      // start the timer,
      start = Call_clk::now();

      // and let's go!
      conn.perform();

      // if we're here, the DL connection has apparently finished successfully
      cout << "\n\n   done\n";

    } catch (curl_easy_exception const& e) {  // oh noes!11  did it broke?
      rprt_throw(e, "Dollnet::dl_hdr_body()");
    }

    // make sure everything's written out for header data
    header_dat << flush;

    // we'll send any DL header data we obtained back to caller
    return header_dat;

  } else
    only_master_can(master(), tag, t(), "Dollnet::dl_hdr_body()");

  ///
  return stringstream{};  // shouldn't reach here
}

//------------------------------------------------------------------------------
auto Dolldrop_b::Dollnet::get_curl() -> curl_easy {
  return curl_easy{};
}

}  // namespace rw::dollhouse::net

#endif  // __has_include(<curl_easy.h>)

//    --=^=--

// "This apartment has ultra-high bandwidth illegal even to large corporations?"

// rw_dollhouse_lib.v220521
// Copyright (2022)
// License MIT (Expat) https://opensource.org/licenses/MIT
